from __future__ import annotations

from abc import abstractmethod
from typing import Any, cast

from meiga import AnyResult, NotImplementedMethodError

from petisco.base.application.application_info import ApplicationInfo
from petisco.base.application.controller.error_map import ErrorMap
from petisco.base.application.controller.meta_controller import MetaController
from petisco.base.domain.errors.default_http_error_map import DEFAULT_HTTP_ERROR_MAP
from petisco.base.misc.result_mapper import ResultMapper, default_failure_handler


def get_shared_and_default_error_map() -> ErrorMap:
    # This mapping will give preference to shared_error_map over DEFAULT_HTTP_ERROR_MAP
    shared_and_default_error_map = {
        **DEFAULT_HTTP_ERROR_MAP,
        **ApplicationInfo().shared_error_map,
    }
    return shared_and_default_error_map


class Controller(metaclass=MetaController):
    """
    A base class for creating controllers.
    Inherit from this class to convert to domain the request values, configure middlewares and instantiate and execute
     a UseCase.
    """

    @staticmethod
    def get_default_mapper() -> ResultMapper:
        error_map = get_shared_and_default_error_map()
        return ResultMapper(error_map=error_map)

    @staticmethod
    def get_config_mapper(config: dict[str, Any]) -> ResultMapper:
        shared_and_default_error_map = get_shared_and_default_error_map()
        controller_error_map = cast(ErrorMap, getattr(config, "error_map", dict()))

        # This merged_dependencies will give preference to controller_error_map (given in the controller) over
        # shared_and_default_error_map
        error_map = {
            **shared_and_default_error_map,
            **controller_error_map,
        }

        return ResultMapper(
            error_map=error_map,
            success_handler=getattr(config, "success_handler", lambda result: result),
            failure_handler=getattr(config, "failure_handler", default_failure_handler),
        )

    @abstractmethod
    def execute(self, *args: Any, **kwargs: Any) -> AnyResult:
        return NotImplementedMethodError



from abc import ABC, abstractmethod
from functools import wraps
from inspect import iscoroutinefunction, signature
from types import FunctionType
from typing import Any, Callable, Dict, Tuple

import elasticapm
from meiga import AnyResult, Error, Failure, NotImplementedMethodError

from petisco.base.application.use_case.use_case_uncontrolled_error import (
    UseCaseUncontrolledError,
)
from petisco.extra.meiga import WaitingForEarlyReturn


def use_case_wrapper(
    method: Callable[..., AnyResult], wrapped_class_name: str
) -> Callable[..., AnyResult]:
    @wraps(method)
    def wrapped(*args: Any, **kwargs: Any) -> AnyResult:
        try:
            return method(*args, **kwargs)
        except WaitingForEarlyReturn as exception:
            return exception.result
        except Error as error:
            return Failure(error)
        except Exception as exception:
            arguments = signature(method).bind(*args, **kwargs).arguments
            arguments.pop("self")
            uncontrolled_error = UseCaseUncontrolledError.from_exception(
                exception=exception, arguments=arguments, class_name=wrapped_class_name
            )
            client = elasticapm.get_client()
            if client:
                client.capture_exception()

            return Failure(uncontrolled_error)

    return wrapped


def async_use_case_wrapper(
    method: Callable[..., AnyResult], wrapped_class_name: str
) -> Callable[..., AnyResult]:
    @wraps(method)
    async def wrapped(*args: Any, **kwargs: Any) -> AnyResult:
        try:
            return await method(*args, **kwargs)
        except WaitingForEarlyReturn as exception:
            return exception.result
        except Error as error:
            return Failure(error)
        except Exception as exception:
            arguments = signature(method).bind(*args, **kwargs).arguments
            arguments.pop("self")
            uncontrolled_error = UseCaseUncontrolledError.from_exception(
                exception=exception, arguments=arguments, class_name=wrapped_class_name
            )
            client = elasticapm.get_client()
            if client:
                client.capture_exception()

            return Failure(uncontrolled_error)

    return wrapped


class MetaUseCase(type, ABC):
    def __new__(
        mcs, name: str, bases: Tuple[Any], namespace: Dict[str, Any]
    ) -> "MetaUseCase":
        new_class_dict = {}
        if "execute" not in namespace:
            raise NotImplementedError(
                "Petisco UseCase or AsyncUseCase must implement an execute method"
            )

        for attributeName, attribute in namespace.items():
            if isinstance(attribute, FunctionType) and attribute.__name__ == "execute":
                if iscoroutinefunction(attribute):
                    attribute = async_use_case_wrapper(attribute, name)
                else:
                    attribute = use_case_wrapper(attribute, name)
            new_class_dict[attributeName] = attribute
        return type.__new__(mcs, name, bases, new_class_dict)

    @abstractmethod
    def execute(self, *args: Any, **kwargs: Any) -> AnyResult:
        return NotImplementedMethodError



from __future__ import annotations

from abc import ABC
from collections import defaultdict
from typing import Any, TypeVar

from petisco.base.application.dependency_injection.dependency import Dependency
from petisco.base.misc.singleton import Singleton

T = TypeVar("T", bound=ABC)


class Container(metaclass=Singleton):
    """
    Singleton which contains set dependencies (List[Dependency)) prepared to be instantiated in order to be injected
     in the UseCases of our application
    """

    def __init__(self) -> None:
        self.dependencies: dict[str, Dependency[Any]] = defaultdict()

    @staticmethod
    def get(base_type: type[T], *, alias: str | None = None) -> T:
        """
        Returns an instance of set Dependency.
        """

        container = Container()

        key = (
            base_type.__name__
            if not alias
            else f'{base_type.__name__} (alias="{alias}")'
        )
        dependency = container.dependencies.get(key)
        if dependency is None:
            raise IndexError(
                f"Invalid dependency. `{key}` is not found within available dependencies [{container.get_available_dependencies()}]"
            )
        instance = dependency.get_instance()
        return instance

    @staticmethod
    def set_dependencies(
        dependencies: list[Dependency[Any]] | None = None, overwrite: bool = False
    ) -> None:
        """
        Set dependencies from a list of them.
        """
        if dependencies is None:
            dependencies = []
        Container()._set_dependencies(dependencies, overwrite)

    @staticmethod
    def get_available_dependencies() -> list[str]:
        """
        Returns the names (keys) of set dependencies.
        """
        keys = list(Container().dependencies.keys())
        return keys

    def _set_dependencies(
        self, input_dependencies: list[Dependency[Any]], overwrite: bool = False
    ) -> None:
        for dependency in input_dependencies:
            key = dependency.get_key()

            if key in self.dependencies and not overwrite:
                if dependency.alias:
                    raise IndexError(
                        f"Container: dependency ({dependency.type.__name__} with alias={dependency.alias}) is already added to dependencies. Check "
                        f"set_dependencies input "
                    )
                else:
                    raise IndexError(
                        f"Container: dependency (type={dependency.type.__name__}) is already added to "
                        f"dependencies. Use Dependency alias to set different dependencies with the same base type "
                    )

            self.dependencies[key] = dependency



from datetime import datetime

from pydantic import PlainSerializer

from petisco.base.domain.message.message import TIME_FORMAT


def serialize_datetime(value: datetime) -> str:
    return value.strftime(TIME_FORMAT)


DatetimeSerializer = PlainSerializer(serialize_datetime)



from __future__ import annotations

from abc import abstractmethod
from typing import Generic, TypeVar

from petisco.base.domain.persistence.database import Database

T = TypeVar("T")


class AsyncDatabase(Generic[T], Database[T]):
    @abstractmethod
    async def is_available(self) -> bool:
        raise NotImplementedError



import os
from functools import wraps
from inspect import signature
from typing import Any, Callable, Dict, List, Type, Union

import elasticapm
from loguru import logger
from meiga import Error, Failure, Result

from petisco.base.application.middleware.middleware import Middleware
from petisco.base.application.middleware.notifier_middleware import NotifierMiddleware
from petisco.base.application.middleware.print_middleware import PrintMiddleware
from petisco.base.application.middleware.request_responded_middleware import (
    RequestRespondedMiddleware,
)
from petisco.base.domain.errors.critical_error import CriticalError
from petisco.base.domain.errors.unknown_error import UnknownError
from petisco.base.domain.value_objects.operation_type import OperationType
from petisco.base.misc.result_mapper import ResultMapper
from petisco.extra.meiga import WaitingForEarlyReturn


def get_middleware_instances(config: Dict[str, Any]) -> List[Middleware]:
    middlewares_instances: List[Middleware] = []

    middlewares_configs = getattr(config, "middlewares", [])
    if not middlewares_configs:
        middlewares_configs = get_middlewares_configuration_from_environment()

    for middlewares_config in middlewares_configs:
        if not isinstance(middlewares_config, Middleware):
            middlewares_instances.append(middlewares_config())
        else:
            middlewares_instances.append(middlewares_config)

    return middlewares_instances


def get_middlewares_configuration_from_environment():
    def gettype(name: str) -> Type[Middleware]:
        lookup_table = {
            "NotifierMiddleware": NotifierMiddleware,
            "RequestRespondedMiddleware": RequestRespondedMiddleware,
            "PrintMiddleware": PrintMiddleware,
        }
        my_type = lookup_table.get(name)
        if not my_type:
            raise TypeError(
                f"Value {name} in PETISCO_DEFAULT_MIDDLEWARES is not valid. Please, use {list(lookup_table.keys())}"
            )
        return my_type

    default_middlewares_names: Union[str, None] = os.getenv(
        "PETISCO_DEFAULT_MIDDLEWARES"
    )
    if not default_middlewares_names:
        return []
    default_middlewares_names_list = default_middlewares_names.split(",")
    return [
        gettype(middleware_name) for middleware_name in default_middlewares_names_list
    ]


def wrapper(
    execute_func: Callable[..., Any],
    wrapped_class_name: str,
    config: Any,
    mapper: ResultMapper,
) -> Callable[..., Any]:
    @wraps(execute_func)
    def wrapped(*args: Any, **kwargs: Any) -> Any:
        from petisco import Controller, MessageSubscriber

        middlewares = get_middleware_instances(config)
        arguments = signature(execute_func).bind(*args, **kwargs).arguments
        arguments.pop("self")

        for middleware in middlewares:
            if (
                middleware.operation_affected == OperationType.ALL
                or middleware.operation_affected == OperationType.CONTROLLER
                and issubclass(args[0].__class__, Controller)
                or middleware.operation_affected == OperationType.SUBSCRIBER
                and issubclass(args[0].__class__, MessageSubscriber)
            ):
                pass
            else:
                continue

            try:
                middleware.set_data(wrapped_class_name, arguments)
                middleware.before()
            except Exception as exception:
                logger.error(f"Error in the {wrapped_class_name} middlewares (before).")
                logger.exception(exception)

        try:
            result = execute_func(*args, **kwargs)
        except WaitingForEarlyReturn as exc:
            result = exc.result
        except Error as error:
            result = Failure(error)
        except Exception as exception:
            unknown_error = UnknownError.from_exception(
                exception=exception,
                arguments=arguments if len(arguments) > 0 else args,
                class_name=wrapped_class_name,
            )
            result = Failure(unknown_error)
            client = elasticapm.get_client()
            if client:
                client.capture_exception()

        if not isinstance(result, Result):
            raise TypeError(
                f"Controller Error: Return value `{result}` ({type(result)}) must be a `meiga.Result`"
            )

        if isinstance(result.value, CriticalError) and len(arguments) > 0:
            formatted_args = {
                k: str(v) if type(v) is not bytes else "bytes"
                for k, v in arguments.items()
            }
            result.value.set_additional_info(formatted_args)

        result.set_transformer(mapper.map)
        for middleware in middlewares:
            if (
                middleware.operation_affected == OperationType.ALL
                or middleware.operation_affected == OperationType.CONTROLLER
                and issubclass(args[0].__class__, Controller)
                or middleware.operation_affected == OperationType.SUBSCRIBER
                and issubclass(args[0].__class__, MessageSubscriber)
            ):
                pass
            else:
                continue

            if result:
                try:
                    middleware.after(result)
                except Exception as exception:
                    logger.error(
                        f"Error in the {wrapped_class_name} middlewares (after)."
                    )
                    logger.exception(exception)

        return result

    return wrapped



from meiga import AnyResult

from petisco import __version__
from petisco.base.application.application_info import ApplicationInfo
from petisco.base.application.dependency_injection.container import Container
from petisco.base.application.middleware.middleware import Middleware
from petisco.base.application.notifier.notifier import Notifier
from petisco.base.application.notifier.notifier_exception_message import (
    NotifierExceptionMessage,
)
from petisco.base.application.notifier.notifier_message import NotifierMessage
from petisco.base.domain.errors.critical_error import CriticalError
from petisco.base.domain.errors.unknown_error import UnknownError


class NotifierMiddleware(Middleware):
    """
    Middleware Implementation to notify critical and unknown errors.

    This Middleware will check result and notify if necessary using Container.get(Notifier) set dependency.
    """

    def __init__(self) -> None:
        self.notifier = Container.get(Notifier)

    def before(self) -> None:
        pass

    def after(self, result: AnyResult) -> None:
        if result.is_failure:
            error = result.value

            app_meta = {
                "petisco": __version__,
                "application_name": ApplicationInfo().name,
                "application_version": ApplicationInfo().version,
            }
            input_meta = self.get_meta_from_input()

            meta = {**app_meta, **input_meta}

            if issubclass(error.__class__, UnknownError):
                notifier_exception_message = (
                    NotifierExceptionMessage.from_unknown_error(
                        error, title="Uncontrolled Exception"
                    )
                )
                notifier_exception_message.update_meta(meta)
                self.notifier.publish_exception(notifier_exception_message)
            if issubclass(error.__class__, CriticalError):
                notifier_message = NotifierMessage(
                    title=error.get_specific_detail(),
                    message=error.__repr__(),
                    meta=meta,
                )
                self.notifier.publish(notifier_message)



from __future__ import annotations

import argparse
import importlib.util
import inspect
import sys
from typing import Any

from loguru import logger
from sqlalchemy import inspect as sql_inspect


def has_args(args: Any) -> bool:
    is_active = False
    for arg in vars(args):
        is_active = is_active or getattr(args, arg)
    return is_active


def get_application(module_path: str):
    logger.remove()
    logger.add(sys.stderr, level="ERROR")

    try:
        module_name = importlib.import_module(module_path)
        application = getattr(module_name, "application")
        return application
    except Exception as exc:  # noqa
        print(
            f"Error loading the app module. Impossible to extract information from {module_path}. "
            f"There is no application instance in {module_path}"
        )
        print(
            "Try to install your app module with `pip install -e .` (maybe you need to add a setup.py or a "
            "pyproject.toml to your app root"
        )
        if is_rich_available():
            from rich.console import Console

            console = Console()
            console.print_exception()
        else:
            print(str(exc))


def encourage_rich_installation() -> None:
    print(
        "\n⚠️ If you want to represent this info in a fancy way. Try to install rich or just install petisco[rich]"
    )


def is_rich_available() -> bool:
    try:
        import rich  # noqa
    except (RuntimeError, ImportError):
        return False
    return True


def show_info(application) -> None:
    dependencies = application.dependencies_provider()
    configurers = application.configurers

    if is_rich_available() is False:
        print("Petisco Application:")
        print(f"name: {application.name}")
        print(f"version: {application.version}")
        print(f"organization: {application.organization}")
        print(f"deployed_at: {application.deployed_at}")
        print(f"dependencies: {len(dependencies)}")
        print(f"configurers: {len(configurers)}")
        encourage_rich_installation()
    else:
        from rich import print
        from rich.panel import Panel

        panel = Panel.fit(
            f"[bold]name:         [/] {application.name}\n"
            f"[bold]version:      [/] {application.version}\n"
            f"[bold]organization: [/] {application.organization}\n"
            f"[bold]deployed_at:  [/] {application.deployed_at}\n"
            f"[bold]dependencies: [/] {len(dependencies)}\n"
            f"[bold]configurers:  [/] {len(configurers)}",
            title="Petisco Application",
        )
        print(panel)


def show_dependencies(application) -> None:
    dependencies = application.dependencies_provider()

    if is_rich_available() is False:
        print("Dependencies:")
        for dependency in dependencies:
            text = f"{dependency.type.__name__} -> "
            for key, builder in dependency.builders.items():
                text += f"{key}: {builder.klass.__name__} | "
            text = text[:-2]
            print(text)
        encourage_rich_installation()
    else:
        from rich.console import Console
        from rich.table import Table

        table = Table()
        console = Console()

        table.add_column("Type", justify="right", style="cyan", no_wrap=True)
        table.add_column("Default")
        table.add_column("Implementations", style="green")
        table.add_column("ENV", style="red")

        for dependency in dependencies:
            type = dependency.get_key()
            default_implementation = dependency.builders.pop("default").klass.__name__

            implementations = ""
            for key, builder in dependency.builders.items():
                implementations += f"{key}: {builder.klass.__name__}\n"
            implementations = implementations[:-1]

            table.add_row(
                type, default_implementation, implementations, dependency.envar_modifier
            )

        console.print(table)


def show_configurers(application) -> None:
    configurers = application.configurers

    if is_rich_available() is False:
        print("Configurers:")
        for configurer in configurers:
            print(
                f"{configurer.__class__.__name__} -> {configurer.execute_after_dependencies=}"
            )
        encourage_rich_installation()
    else:
        from rich.console import Console
        from rich.table import Table

        table = Table()
        console = Console()

        table.add_column("Configurer", justify="right", style="cyan", no_wrap=True)
        table.add_column("execute_after_dependencies", justify="center")

        for configurer in configurers:
            table.add_row(
                configurer.__class__.__name__,
                str(configurer.execute_after_dependencies),
            )

        console.print(table)


def show_sql_models(declarative_base_path: str) -> None:
    module_name, class_name = declarative_base_path.rsplit(".", 1)
    module = importlib.import_module(module_name)
    Base = getattr(module, class_name)

    info: list[dict[str, str]] = []
    for sql_model in Base.__subclasses__():
        info.append(
            {
                "name": sql_model.__name__,
                "module": sql_model.__module__,
                "filename": inspect.getsourcefile(sql_model),
            }
        )

    if is_rich_available() is False:
        print(info)
    else:
        from rich.console import Console
        from rich.table import Table

        table = Table()
        console = Console()

        table.add_column("Model", justify="right", style="cyan", no_wrap=True)
        table.add_column("Filename", justify="left")

        for model_info in info:
            filename = model_info.get("filename")
            table.add_row(
                model_info.get("name"),
                f"file://{filename}",
            )
        console.print(table)


def generate_sql_models_diagram(declarative_base_path: str) -> None:
    try:
        import graphviz  # noqa
    except (RuntimeError, ImportError):
        print(
            "\n⚠️ If you want to generate a SVG diagram of Petisco SQL models you need to install graphviz. Try to install graphviz or just install petisco[dev]"
        )
        return

    module_name, class_name = declarative_base_path.rsplit(".", 1)
    module = importlib.import_module(module_name)
    Base = getattr(module, class_name)

    models: list[tuple] = []
    for sql_model in Base.__subclasses__():
        models.append((sql_model.__module__, sql_model.__name__))

    try:
        generate_data_model_diagram(models)
    except graphviz.ExecutableNotFound as exc:
        print(exc)
        print(
            "\n⚠️ The Graphviz Python package has already been installed, but it requires an executable to be available.\n🍎 If you're using macOS, please ensure that the Graphviz executable is installed. You can do this by running 'brew install graphviz' in your terminal."
        )
    except Exception as exc:
        raise exc


def generate_data_model_diagram(
    models: list[tuple], output_file: str = "petisco_sql_models_diagram.svg"
) -> None:
    import graphviz

    # Initialize graph with more advanced visual settings
    dot = graphviz.Digraph(
        comment="Interactive Data Models",
        format="svg",
        graph_attr={"bgcolor": "#EEEEEE", "rankdir": "TB", "splines": "spline"},
        node_attr={"shape": "none", "fontsize": "12", "fontname": "Roboto"},
        edge_attr={"fontsize": "10", "fontname": "Roboto"},
    )

    # Iterate through each SQLAlchemy model
    for model in models:
        module = importlib.import_module(model[0])
        model_instance = getattr(module, model[1])
        insp = sql_inspect(model_instance)
        name = insp.class_.__name__

        # Create an HTML-like label for each model as a rich table
        label = f"""<
        <TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0">
        <TR><TD COLSPAN="2" BGCOLOR="#3F51B5"><FONT COLOR="white">{name}</FONT></TD></TR>
        """

        for column in insp.columns:
            constraints = []
            if column.primary_key:
                constraints.append("PK")
            if column.unique:
                constraints.append("Unique")
            if column.index:
                constraints.append("Index")

            constraint_str = ",".join(constraints)
            color = "#BBDEFB"

            label += f"""<TR>
                         <TD BGCOLOR="{color}">{column.name}</TD>
                         <TD BGCOLOR="{color}">{column.type} ({constraint_str})</TD>
                         </TR>"""

        label += "</TABLE>>"

        # Create the node with added hyperlink to detailed documentation
        dot.node(name, label=label, URL=f"http://{name}_details.html")

        # Add relationships with tooltips and advanced styling
        for rel in insp.relationships:
            target_name = rel.mapper.class_.__name__
            tooltip = f"Relation between {name} and {target_name}"
            dot.edge(
                name,
                target_name,
                label=rel.key,
                tooltip=tooltip,
                color="#1E88E5",
                style="dashed",
            )

    dot.render(output_file, view=True)


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="petisco-dev 🍪",
        description="petisco dev tools to inspect your application",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-i", "--info", action="store_true", help="show petisco app info."
    )
    parser.add_argument(
        "-deps",
        "--dependencies",
        action="store_true",
        help="show petisco app dependencies.",
    )
    parser.add_argument(
        "-configs",
        "--configurers",
        action="store_true",
        help="show petisco app configurers.",
    )
    parser.add_argument(
        "-sql-models",
        "--sql-models",
        action="store_true",
        help="show petisco sql models.",
    )
    parser.add_argument(
        "--sql-models-diagram",
        "--sql-models-diagram",
        action="store_true",
        help="generate petisco SQL models diagram.",
    )
    parser.add_argument(
        "-declarative-base",
        "--declarative-base",
        default="petisco.extra.sqlalchemy.SqlBase",
        help="path to DeclarativeBase, a class to gather all the SQL models",
    )
    parser.add_argument(
        "--application",
        default="app.application",
        help="Module path (default app.application)",
    )

    args = parser.parse_args()

    if not has_args(args):
        parser.print_help()
    else:
        application = get_application(args.application)
        if application is None:
            return

        if args.info:
            show_info(application)
            return

        if args.dependencies:
            show_dependencies(application)
            return

        if args.configurers:
            show_configurers(application)
            return

        if args.sql_models:
            show_sql_models(args.declarative_base)
            return

        if args.sql_models_diagram:
            generate_sql_models_diagram(args.declarative_base)
            return



import meiga

if meiga.__version__ < "1.9.4":
    from meiga.on_failure_exception import (  # noqa
        OnFailureException as WaitingForEarlyReturn,
    )
elif meiga.__version__ == "1.9.4":
    from meiga.failures import WaitingForEarlyReturn  # noqa
else:
    from meiga import WaitingForEarlyReturn  # noqa



import os
from typing import Any, Union
from uuid import UUID, uuid4

import validators
from pydantic import GetCoreSchemaHandler, GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema

from petisco.base.domain.errors.defaults.invalid_uuid import InvalidUuid

USE_LEGACY_UUID = bool(os.getenv("USE_LEGACY_UUID", "False").lower() == "true")


class Uuid(str):
    def __new__(cls, value: Union[str, UUID]) -> "Uuid":
        if value is None or not validators.uuid(str(value)):
            raise InvalidUuid(uuid_value=str(value))
        return super().__new__(cls, str(value))

    @classmethod
    def v4(cls) -> "Uuid":
        return cls(str(uuid4()))

    @property
    def value(self) -> str:
        return self

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        """
        We return a pydantic_core.CoreSchema that behaves in the following ways:

        * str will be parsed as `Uuid` instances with the str as the value attribute
        * `Uuid` instances will be parsed as `Uuid` instances without any changes
        * Nothing else will pass validation
        * Serialization will always return just a str
        """

        def validate_from_str(value: str) -> "Uuid":
            return Uuid(value)

        from_uuid_schema = core_schema.chain_schema(
            [
                core_schema.uuid_schema(),
                core_schema.no_info_plain_validator_function(validate_from_str),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_uuid_schema,
            python_schema=core_schema.union_schema(
                [
                    # check if it's an instance first before doing any further work
                    core_schema.is_instance_schema(Uuid),
                    from_uuid_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda instance: instance.value
            ),
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls, _core_schema: core_schema.CoreSchema, handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        # Use the same schema that would be used for `uuid`
        return handler(core_schema.uuid_schema())


if USE_LEGACY_UUID is True:
    from petisco.base.domain.model.legacy_uuid import LegacyUuid  # noqa

    Uuid = LegacyUuid  # noqa



import json
from json import JSONDecodeError
from typing import Any, Dict, Union

from loguru import logger
from meiga import Result

from petisco.base.application.controller.http_error import HttpError
from petisco.base.domain.errors.domain_error import DomainError
from petisco.base.domain.message.domain_event import DomainEvent

REQUEST_RESPONDED_UNWRAP_ERROR = (
    {
        "error": {
            "type": "RequestRespondedUnwrapError",
            "message": "Cannot unwrap http response from decorated function",
        }
    },
)


class RequestResponded(DomainEvent):
    app_name: str
    app_version: str
    controller: str
    is_success: bool
    http_response: Dict[str, Any]
    elapsed_time: float
    info_id: Union[str, None] = None

    @staticmethod
    def create(
        app_name: str,
        app_version: str,
        controller: str,
        is_success: bool,
        http_response: Dict[str, Any],
        elapsed_time: float,
        info_id: Union[str, None] = None,
    ) -> "RequestResponded":
        app_name = app_name
        app_version = app_version
        controller = controller
        is_success = is_success
        elapsed_time = elapsed_time
        http_response = RequestResponded.set_http_response(is_success, http_response)
        info_id = info_id

        return RequestResponded(
            app_name=app_name,
            app_version=app_version,
            controller=controller,
            is_success=is_success,
            elapsed_time=elapsed_time,
            http_response=http_response,
            info_id=info_id,
        )

    @staticmethod
    def _get_content(
        is_success: bool, response: Union[Dict[str, Any], str]
    ) -> Dict[str, Any]:
        content = response
        if isinstance(content, str):
            try:
                content = json.loads(content)
            except JSONDecodeError:
                pass
            except Exception as e:
                logger.error(e)
                raise e
        content_size = len(str(content))

        if content_size > 300:
            if is_success:
                content = {
                    "message": f"Response OK (Trimmed message: {content[:300]})",
                    "message_size": content_size,
                }
            else:
                error_type = "Error"
                if isinstance(content, dict):
                    error_type = content.get("error", {}).get("type", "Error")
                content = {
                    "error_type": error_type,
                    "message": f"Response Error (Trimmed message: {content.value[:300]})",
                    "message_size": content_size,
                }
        else:
            if isinstance(content, dict):
                content = {"message": json.dumps(content), "message_size": content_size}
            elif isinstance(content, HttpError):
                content = {"message": content.detail, "message_size": content_size}
            elif isinstance(content, Result):
                if isinstance(content.value, DomainError):
                    content = {
                        "message": content.value.detail(),
                        "message_size": content_size,
                    }
                else:
                    content = {"message": content.value, "message_size": content_size}
            else:
                content = {"message": content, "message_size": content_size}
        return content

    @staticmethod
    def set_http_response(is_success: bool, http_response: Any) -> Dict[str, Any]:
        try:
            _http_response = {
                "content": REQUEST_RESPONDED_UNWRAP_ERROR,
                "status_code": 500,
            }
            if is_success:
                if isinstance(http_response, Result):
                    _http_response["content"] = RequestResponded._get_content(
                        is_success, http_response.value
                    )
                else:
                    _http_response["content"] = RequestResponded._get_content(
                        is_success, http_response
                    )
                _http_response["status_code"] = 200
            else:
                if isinstance(http_response, HttpError):
                    _http_response["content"] = RequestResponded._get_content(
                        is_success, http_response.detail
                    )
                    _http_response["status_code"] = http_response.status_code
                else:
                    _http_response["content"] = RequestResponded._get_content(
                        is_success, http_response
                    )
                    _http_response["status_code"] = 500
        except Exception as e:
            logger.error(e)
            raise e
        return _http_response
